import { App } from '@kevisual/router';
import { loadAppInfo, AppInfoConfig, saveAppInfo } from '../lib/app-file.ts';
import { fork } from 'child_process';
import { merge } from 'lodash-es';
import { deleteFileAppInfo } from './load-app.ts';
import { fileIsExist } from '@kevisual/use-config';
import path from 'path';
// 共享
export const existDenpend = [
  'sequelize', // commonjs
  'pg', // commonjs
  '@kevisual/router', // 共享模块
  'ioredis', // commonjs
  'socket.io', // commonjs
  'minio', // commonjs
  'pino', // commonjs
  'pino-pretty', // commonjs
  '@msgpack/msgpack', // commonjs
];
export enum AppType {
  SystemApp = 'system-app',
  MicroApp = 'micro-app',
  GatewayApp = 'gateway-app',
}
export type AppInfo = {
  key: string;
  status?: 'inactive' | 'running' | 'stop' | 'error'; // 运行状态
  version?: string; // 版本
  type?: AppType; // 默认类型
  description?: string; // 描述
  timestamp?: number; // 时间戳, 每次更新更新时间戳
  process?: any; // 进程

  origin?: Record<string, any>; // 原始数据
  entry?: string; // 入口文件
  path?: string; // 文件路径
};
export const onAppShowInfo = (app: AppInfo) => {
  return {
    key: app.key,
    status: app.status,
    type: app.type,
    description: app.description,
    version: app.version,
  };
};
export const createAppShowInfo = (app: any) => {
  return {
    key: app.key,
    status: app.status,
    type: app.type,
    description: app.description,
    version: app.version,
  };
};
type managerOptions = {
  mainApp: App;
};
export class Manager<T extends AppInfo = any> {
  apps: Map<string, T>;
  mainApp: App;
  appInfo: AppInfoConfig;
  constructor(opts: managerOptions) {
    this.apps = new Map();
    this.mainApp = opts.mainApp;
    this.appInfo = {} as any;
  }
  /**
   * 检查key是否存在
   * @param key
   * @returns
   */
  checkKey(key: string) {
    return this.apps.has(key);
  }
  /*
   * 获取app信息
   * @param key
   */
  async loadApp(app: T) {
    const mainApp = this.mainApp;
    this.apps.set(app.key, app);
    if (app.status !== 'running') {
      return;
    }
    if (!fileIsExist(app.path)) {
      console.error('app is not found');
      return;
    }
    const pathEntry = path.join(app.path, app.entry);
    if (!fileIsExist(pathEntry)) {
      console.error('file entry not found');
      return;
    }
    const entry = app.entry + `?timestamp=${app?.timestamp}`;
    // 注册路由
    if (app.type === AppType.MicroApp) {
      const process = fork(app.entry, [], {
        stdio: 'inherit', // 共享主进程的标准输入输出
        cwd: app.path,
      });
      app.process = process;
    } else if (app.type === AppType.SystemApp) {
      const module = await import(entry);
      if (module.loadApp) {
        await module.loadApp?.(mainApp);
      }
    } else if (app.type === AppType.GatewayApp) {
      console.log('gateway app not support');
    }
    return true;
  }
  /**
   * create new app info
   * @param app
   */
  async saveAppInfo(app: T, newTimeData = false) {
    const list = this.appInfo.list || [];
    if (newTimeData) {
      app.timestamp = Date.now();
    }
    const { process, ...info } = app;
    list.push(info);
    this.appInfo.list = list;
    await saveAppInfo(this.appInfo);
  }

  /**
   * 初始化应用的时候加载
   */
  async load() {
    // 从apps文件夹列表当中中加载app信息
    const appInfos = await loadAppInfo();
    this.appInfo = appInfos;
    const list = appInfos?.list || [];
    for (const app of list) {
      try {
        const loaded = await this.loadApp(app);
        if (!loaded) {
          // 加载失败，如果是running状态，设置为error
          if (app.status === 'running') {
            app.status = 'error';
            console.log('load app error', app); // save app error info
            await this.saveAppInfo(app);
          }
        } else {
          // console.log('load app success', app);
        }
      } catch (e) {
        console.error('load app', e);
      }
    }
  }
  async add(app: T) {
    if (this.checkKey(app.key)) {
      console.error('key is loaded');
      return false;
    }

    await this.saveAppInfo(app, true);
    this.loadApp(app);
  }
  // 启动
  async start(key: string) {
    const app = this.apps.get(key);
    if (!app) {
      return;
    }
    if (app.status === 'running') {
      return;
    }
    app.status = 'running';
    this.loadApp(app);
    await this.saveAppInfo(app);
  }
  // 停止
  async stop(key: string) {
    const app = this.apps.get(key);
    if (!app) {
      return;
    }
    if (app.status === 'stop') {
      return;
    }
    app.status = 'stop';
    if (app.process) {
      app.process.kill();
    }
    await this.saveAppInfo(app);
  }
  /**
   * 获取app信息, 用于展示
   * @param key
   * @returns
   */
  getAppShowInfo(key: string) {
    const app = this.apps.get(key);
    if (!app) {
      return;
    }
    return onAppShowInfo(app);
  }
  /**
   * 获取所有app信息, 用于展示
   * @returns
   */
  getAllAppShowInfo() {
    const list = [];
    for (const [key, value] of this.apps) {
      list.push(onAppShowInfo(value));
    }
    return list;
  }
  /**
   * 更新app信息, 用于展示, 加上一些功能，启动，停止程序
   * @param key
   * @param info
   * @returns
   */
  async updateAppInfo(key: string, info: Partial<T>) {
    const app = this.apps.get(key);
    if (!app) {
      return;
    }
    merge(app, info);
    this.loadApp(app);
    await this.saveAppInfo(app);
    return onAppShowInfo(app);
  }
  /**
   * 删除app信息
   * @param key
   * @returns
   */
  async removeApp(key: string) {
    const app = this.apps.get(key);
    if (!app) {
      return;
    }
    if (app.process) {
      app.process.kill();
    }
    this.apps.delete(key);
    await this.saveAppInfo(app);
    try {
      deleteFileAppInfo(key);
    } catch (e) {
      console.error('delete file app error', e);
    }
  }
}
